///
/// Returns the return type of the function type `Fn`.
///
pub fn FnReturn(comptime Fn: type) type {
    const type_info = @typeInfo(Fn);

    if (type_info != .Fn) @compileError("`Fn` must be a function type");

    return type_info.Fn.return_type orelse void;
}

///
/// Returns single-input closure type where `Input` is the input type and `Output` is the output
/// type.
///
pub fn Function(comptime Input: type, comptime Output: type) type {
    return struct {
        context: *anyopaque,
        contextualApply: fn (*anyopaque, Input) Output,

        ///
        /// Function type.
        ///
        const Self = @This();

        ///
        /// Applies `input` to `self`, producing a result according to the type-erased
        /// implementation.
        ///
        pub fn apply(self: Self, input: Input) Output {
            return self.contextualApply(self.context, input);
        }

        ///
        /// Creates a new [Self] by wrapping `concrete_context` as a pointer to the implementation
        /// and `contextualApply` as the behavior executed when [apply] is called.
        ///
        /// The newly created [Self] is returned.
        ///
        pub fn wrap(
            concrete_context: anytype,
            comptime contextualApply: fn (@TypeOf(concrete_context), Input) Output
        ) Self {
            const ConcreteContext = @TypeOf(concrete_context);

            if (@typeInfo(ConcreteContext) != .Pointer)
                @compileError("`concrete_context` must be a pointer type");

            return .{
                .context = concrete_context,

                .contextualApply = struct {
                    fn call(erased_context: *anyopaque, input: Input) Output {
                        return contextualApply(@ptrCast(ConcreteContext, @alignCast(
                            @alignOf(ConcreteContext), erased_context)), input);
                    }
                }.call,
            };
        }
    };
}
